from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404, redirect
from django.views.generic.base import TemplateView

from .models import SkosCollection, SkosCollectionContentObject


class ContentObjectMixin:
    """
    Setup the ContentType and the object used by a view, based on the
    `content_type_id` and the `object_id` arguments passed in the URL.
    """

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)
        self.content_type = get_object_or_404(ContentType, pk=kwargs["content_type_id"])
        self.object = get_object_or_404(
            self.content_type.model_class(), pk=kwargs["object_id"]
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["content_type"] = self.content_type
        context["object"] = self.object
        return context


class CollectionToggle(LoginRequiredMixin, ContentObjectMixin, TemplateView):
    """
    Toggle a collection - if a CollectionObject connecting the requested object
    and collection does not exist, then create it. If it does exist, delete it.
    """

    template_name = "collections/collection_toggle.html"

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)
        self.collection = get_object_or_404(SkosCollection, pk=kwargs["collection"])

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["exists"] = self.created
        context["collection"] = self.collection
        return context

    def get(self, *args, **kwargs):
        scco, self.created = SkosCollectionContentObject.objects.get_or_create(
            collection=self.collection,
            content_type=self.content_type,
            object_id=self.object.id,
        )
        if not self.created:
            scco.delete()
        if redirect_to := self.request.GET.get("to", False):
            return redirect(redirect_to)
        return super().get(*args, **kwargs)


class CollectionObjectCollection(LoginRequiredMixin, ContentObjectMixin, TemplateView):
    """
    Change the requested CollectionObjects collection to point to the collection the
    current collection is in.
    """

    template_name = "collections/collection_object_collection.html"

    def get_context_data(self, *args, **kwargs):
        collectionobject = get_object_or_404(
            SkosCollectionContentObject, pk=kwargs["collectionobject"]
        )
        parent = collectionobject.collection.parent_collection
        collectionobject.collection = parent
        collectionobject.save()
        context = super().get_context_data(*args, **kwargs)
        context["collectionobject"] = collectionobject
        return context


class CollectionSessionToggle(LoginRequiredMixin, TemplateView):
    """
    Toggle the existence of an SkosCollection in the `session_collections`
    session variable.
    This can be used in combination with the
    `collections.signals.add_to_session_collection` signal, to add objects
    to a collection if the collections id is listed in the session variable.
    The equivalent templateatag that calls this view is
    `collections.templatetags.collections.collection_session_toggle_by_id`
    """

    template_name = "collections/collection_session_toggle.html"

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx["collection"] = self.skoscollection
        ctx["enabled"] = self.skoscollection.id in self.session_collections
        return ctx

    def get(self, *args, **kwargs):
        self.skoscollection = get_object_or_404(
            SkosCollection, pk=kwargs["skoscollection"]
        )
        self.session_collections = set(
            self.request.session.get("session_collections", [])
        )
        self.session_collections ^= {self.skoscollection.id}
        self.request.session["session_collections"] = list(self.session_collections)
        if redirect_to := self.request.GET.get("to", False):
            return redirect(redirect_to)
        return super().get(*args, **kwargs)



# SPDX-FileCopyrightText: 2023 Birger Schacht
# SPDX-License-Identifier: MIT

from pathlib import Path

from django.test import TestCase

from apis_core.utils import rdf

# use `curl -H "Accept: application/rdf+xml" -L $URI` to fetch data

testdata = Path(__file__).parent / "testdata"


class RdfTest(TestCase):
    def test_get_definition_from_dict_place_from_geonames(self):
        achensee = {
            "latitude": ["47.5"],
            "longitude": ["11.7"],
            "label": ["Achensee"],
        }
        # https://www.geonames.org/2783029/achensee.html
        uri = str(testdata / "achensee.rdf")

        attributes = rdf.get_something_from_uri(uri)
        self.assertEqual(achensee["latitude"], attributes["latitude"])
        self.assertEqual(achensee["longitude"], attributes["longitude"])
        self.assertEqual(achensee["label"], attributes["label"])

    def test_get_definition_from_dict_place_from_dnb(self):
        wien = {
            "label": ["Wien"],
            "latitude": ["+048.208199"],
            "longitude": ["016.371690"],
        }
        # https://d-nb.info/gnd/4066009-6
        uri = str(testdata / "wien.rdf")

        attributes = rdf.get_something_from_uri(uri)
        self.assertEqual(wien["latitude"], attributes["latitude"])
        self.assertEqual(wien["longitude"], attributes["longitude"])
        self.assertEqual(wien["label"], attributes["label"])

    def test_get_definition_from_dict_person_from_dnb(self):
        pierre = {
            "forename": [
                "Pierre",
                "Pʹer",
                "Rudolf",
                "Rodolphe",
                "...",
            ],
            "surname": ["Ramus", "Großmann", "Grossmann", "Grossman", "Libertarian"],
            "alternative_names": [
                "Ramus, Pʹer",
                "Großmann, Rudolf",
                "Grossmann, Rudolf",
                "Grossman, Rudolf",
                "Grossman, Rodolphe",
                "Grossmann, Rodolphe",
                "Libertarian, ...",
            ],
            "date_of_birth": ["1882-04-15"],
            "date_of_death": ["1942"],
        }
        # https://d-nb.info/gnd/118833197
        uri = str(testdata / "ramus.rdf")

        attributes = rdf.get_something_from_uri(uri)
        self.assertEqual(pierre["forename"], attributes["forename"])
        self.assertEqual(pierre["surname"], attributes["surname"])
        # self.assertEqual(pierre["alternative_names"], attributes["alternative_names"])
        self.assertEqual(pierre["date_of_birth"], attributes["date_of_birth"])
        self.assertEqual(pierre["date_of_death"], attributes["date_of_death"])

    def test_get_definition_from_dict_institution_from_dnb(self):
        pierre_ges = {
            "label": [
                "Pierre-Ramus-Gesellschaft",
                "Ramus-Gesellschaft",
                "Pierre Ramus-Gesellschaft",
            ],
        }
        # https://d-nb.info/gnd/415006-5
        uri = str(testdata / "ramus_gesellschaft.rdf")

        attributes = rdf.get_something_from_uri(uri)
        self.assertEqual(pierre_ges["label"], attributes["label"])

    def test_get_definition_from_dict_institution_from_dnb2(self):
        oeaw = {
            "label": [
                "Akademie der Wissenschaften in Wien",
                "Academy of Sciences in Vienna",
                "Akademie der Wissenschaften (Wien)",
            ],
        }
        # https://d-nb.info/gnd/35077-1
        uri = str(testdata / "oeaw.rdf")

        attributes = rdf.get_something_from_uri(uri)
        self.assertEqual(oeaw["label"], attributes["label"])



from django.contrib import admin
from django.urls import include, path

urlpatterns = [
    path("", include("apis_core.urls", namespace="apis")),
    # https://docs.djangoproject.com/en/5.0/topics/auth/default/#module-django.contrib.auth.views
    path("accounts/", include("django.contrib.auth.urls")),
    # https://docs.djangoproject.com/en/5.0/ref/contrib/admin/#hooking-adminsite-to-urlconf
    path("admin/", admin.site.urls),
]



from django import template
from django.contrib.contenttypes.models import ContentType

from apis_core.apis_entities.models import AbstractEntity

register = template.Library()


@register.simple_tag
def url_replace(request, field, value):
    dict_ = request.GET.copy()
    dict_[field] = value

    return dict_.urlencode()


def is_entity(content_type: ContentType):
    model_class = content_type.model_class()
    return model_class is not None and issubclass(model_class, AbstractEntity)


@register.simple_tag
def entities_content_types():
    """
    Retrieve all models which inherit from AbstractEntity class
    and return their ContentType.
    """
    entities = list(
        filter(
            lambda content_type: is_entity(content_type),
            ContentType.objects.all(),
        )
    )
    return entities



# Generated by Django 5.1.7 on 2025-04-16 09:25

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("collections", "0003_skoscollection_unique_name_parent_and_more"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="skoscollection",
            name="check_name_pipe",
        ),
        migrations.RemoveField(
            model_name="skoscollection",
            name="parent",
        ),
    ]



import json

from drf_spectacular.utils import OpenApiParameter, extend_schema, inline_serializer
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from apis_core.utils.helpers import datadump_serializer


class Dumpdata(APIView):
    """
    provide an API endpoint that outputs the datadump of an APIS installation

    this is a bit of a hack, becaus we first use the Django JSON serializer to
    serialize the data using natural keys, then we use json.loads to so we can
    output it as an API reponse.
    so basically: serialize -> deserialize -> serialize
    """

    permission_classes = [IsAuthenticated]

    @extend_schema(
        parameters=[OpenApiParameter(name="app_labels", type=str, many=True)],
        responses={200: inline_serializer(name="DumpDataResponse", fields={})},
    )
    def get(self, request, *args, **kwargs):
        params = request.query_params.dict()
        app_labels = params.pop("app_labels", [])
        if app_labels:
            app_labels = app_labels.split(",")
        return Response(json.loads(datadump_serializer(app_labels, "json")))



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured

from apis_core.apis_metainfo.models import Uri
from apis_core.utils.helpers import get_importer_for_model


def create_object_from_uri(uri: str, model: object, raise_on_fail=False) -> object:
    if uri.startswith("http"):
        try:
            uri = Uri.objects.get(uri=uri)
            return uri.content_object
        except Uri.DoesNotExist:
            Importer = get_importer_for_model(model)
            importer = Importer(uri, model)
            instance = importer.create_instance()
            content_type = ContentType.objects.get_for_model(instance)
            uri = Uri.objects.create(
                uri=importer.get_uri,
                content_type=content_type,
                object_id=instance.id,
            )
            return instance
    if raise_on_fail:
        content_type = ContentType.objects.get_for_model(model)
        raise ImproperlyConfigured(
            f'Could not create {content_type.name} from string "{uri}"'
        )
    return False



from django.apps import apps
from django.contrib.contenttypes.models import ContentType

from apis_core.apis_entities.models import AbstractEntity


def get_entity_classes():
    return list(filter(lambda x: issubclass(x, AbstractEntity), apps.get_models()))


def get_entity_content_types():
    return [ContentType.objects.get_for_model(model) for model in get_entity_classes()]



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django_filters import CharFilter, MultipleChoiceFilter

from apis_core.apis_metainfo.models import RootObject
from apis_core.generic.filtersets import GenericFilterSet
from apis_core.generic.helpers import generate_search_filter
from apis_core.relations.forms import RelationFilterSetForm
from apis_core.relations.utils import get_all_relation_subj_and_obj


class EntityFilter(CharFilter):
    """
    Custom CharFilter that uses the generate_search_filter helper
    to search in all instances inheriting from RootObject and then
    uses those results to only list relations that point to one of
    the results.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extra["help_text"] = "Searches in subclasses of RootObject"

    def _search_all_entities(self, value) -> list[str]:
        q = Q()
        for content_type in get_all_relation_subj_and_obj():
            name = content_type.model
            q |= Q(**{f"{name}__isnull": False}) & generate_search_filter(
                content_type.model_class(), value, prefix=f"{name}__"
            )
        return RootObject.objects_inheritance.filter(q).values_list("pk", flat=True)

    def filter(self, qs, value):
        if value:
            all_entities = self._search_all_entities(value)
            return qs.filter(**{f"{self.field_name}_object_id__in": all_entities})
        return qs


class SubjObjClassFilter(MultipleChoiceFilter):
    """
    Custom MultipleChoiceFilter
    * it lists model classes as choices, that are in some way
      connected to a relation, either as subj or as obj
    * it filters relations by the content types of the connected
      subjects and objects
    """

    def __init__(self, models, *args, **kwargs):
        super().__init__(*args, **kwargs)
        content_types = [ContentType.objects.get_for_model(model) for model in models]
        self.extra["choices"] = [(item.id, item.name) for item in content_types]

    def filter(self, qs, value: list[str] | None):
        # value is the list of contenttypes ids
        if value:
            return qs.filter(Q(**{f"{self.field_name}_content_type__in": value}))
        return qs


class RelationFilterSet(GenericFilterSet):
    """
    Override the GenericFilterSet that is created for the Relation model.
    It does not really make sense to filter for content type id and object id,
    so we exclude those.
    Instead, we add a multiple choice filter for object and subject class, that
    only lists those choices that actually exists (meaning the classes that are
    actually set as subj or obj in some relation).
    Additionaly, we add a search filter, that searches in instances connected
    to relations (this does only work for instances inheriting from RootObject).
    """

    subj_search = EntityFilter(
        field_name="subj",
        label="Subject search",
    )
    obj_search = EntityFilter(
        field_name="obj",
        label="Object search",
    )

    class Meta:
        exclude = [
            "subj_object_id",
            "subj_content_type",
            "obj_object_id",
            "obj_content_type",
        ]
        form = RelationFilterSetForm

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if model := getattr(self.Meta, "model", False):
            all_models = [ct.model_class() for ct in get_all_relation_subj_and_obj()]
            subj_models = getattr(model, "subj_model", all_models)
            obj_models = getattr(model, "obj_model", all_models)
            if isinstance(subj_models, list):
                self.filters["subj_class"] = SubjObjClassFilter(
                    field_name="subj", label="Subject class", models=subj_models
                )
            if isinstance(obj_models, list):
                self.filters["obj_class"] = SubjObjClassFilter(
                    field_name="obj", label="Object Class", models=obj_models
                )



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from apis_core.generic.abc import GenericModel


class SkosCollectionManager(models.Manager):
    def by_instance(self, instance):
        content_type = ContentType.objects.get_for_model(instance)
        scco = SkosCollectionContentObject.objects.filter(
            content_type=content_type, object_id=instance.id
        )
        return self.get_queryset().filter(
            pk__in=scco.values_list("collection", flat=True)
        )


class SkosCollection(GenericModel, models.Model):
    """
    SKOS collections are labeled and/or ordered groups of SKOS concepts.
    Collections are useful where a group of concepts shares something in common,
    and it is convenient to group them under a common label, or
    where some concepts can be placed in a meaningful order.

    Miles, Alistair, and Sean Bechhofer. "SKOS simple knowledge
    organization system reference. W3C recommendation (2009)."

    """

    class Meta:
        ordering = ["name"]

    name = models.CharField(
        max_length=300,
        verbose_name="skos:prefLabel",
        help_text="Collection label or name",
    )
    label_lang = models.CharField(
        max_length=3,
        blank=True,
        default="en",
        verbose_name="skos:prefLabel language",
        help_text="Language of preferred label given above",
    )
    creator = models.TextField(
        blank=True,
        verbose_name="dc:creator",
        help_text="Person or organisation that created this collection"
        "If more than one list all using a semicolon ;",
    )
    contributor = models.TextField(
        blank=True,
        verbose_name="dc:contributor",
        help_text="Person or organisation that made contributions to the collection"
        "If more than one list all using a semicolon ;",
    )
    objects = SkosCollectionManager()

    def __str__(self):
        return self.name

    def add(self, instance: object):
        content_type = ContentType.objects.get_for_model(instance)
        SkosCollectionContentObject.objects.get_or_create(
            collection=self, content_type=content_type, object_id=instance.pk
        )

    def remove(self, instance: object):
        content_type = ContentType.objects.get_for_model(instance)
        SkosCollectionContentObject.objects.filter(
            collection=self, content_type=content_type, object_id=instance.pk
        ).delete()

    @property
    def parent_collection(self):
        content_type = ContentType.objects.get_for_model(self)
        sccos = SkosCollectionContentObject.objects.filter(
            content_type=content_type, object_id=self.id
        )
        if len(sccos) == 1:
            return sccos.first().collection
        raise SkosCollection.MultipleObjectsReturned(
            f'"{self}" is part of multiple collections'
        )


class SkosCollectionContentObject(GenericModel, models.Model):
    """
    *Throughtable* datamodel to connect collections to arbitrary content
    """

    collection = models.ForeignKey(SkosCollection, on_delete=models.CASCADE)

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    def __str__(self):
        return f"{self.content_object} -> {self.collection}"



from django.core.management.base import BaseCommand

from apis_core.utils.helpers import datadump_serializer


class Command(BaseCommand):
    help = "Dump APIS data"

    def add_arguments(self, parser):
        parser.add_argument(
            "args",
            metavar="app_labels",
            nargs="*",
            help=("Optional additional app_labels."),
        )

    def handle(self, *app_labels, **options):
        print(datadump_serializer(app_labels, "json"))



from dal_select2.widgets import (
    ListSelect2,
    ModelSelect2,
    ModelSelect2Multiple,
    Select2,
    Select2Multiple,
    TagSelect2,
)


class BootstrapAttrMixin:
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        attrs["data-theme"] = "bootstrap-5"
        return attrs


class ApisListSelect2(BootstrapAttrMixin, ListSelect2):
    autocomplete_function = "apis_select2"


class ApisModelSelect2(BootstrapAttrMixin, ModelSelect2):
    autocomplete_function = "apis_select2"


class ApisModelSelect2Multiple(BootstrapAttrMixin, ModelSelect2Multiple):
    autocomplete_function = "apis_select2"


class ApisSelect2(BootstrapAttrMixin, Select2):
    autocomplete_function = "apis_select2"


class ApisSelect2Multiple(BootstrapAttrMixin, Select2Multiple):
    autocomplete_function = "apis_select2"


class ApisTagSelect2(BootstrapAttrMixin, TagSelect2):
    autocomplete_function = "apis_select2"



import inspect
from typing import Any

import django
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import AppRegistryNotReady
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from simple_history import utils
from simple_history.models import HistoricalRecords

from apis_core.generic.abc import GenericModel


class APISHistoricalRecords(HistoricalRecords, GenericModel):
    def get_m2m_fields_from_model(self, model):
        # Change the original simple history function to also return m2m fields
        m2m_fields = []
        try:
            for field in inspect.getmembers(model):
                if isinstance(
                    field[1],
                    django.db.models.fields.related_descriptors.ManyToManyDescriptor,
                ):
                    m2m_fields.append(getattr(model, field[0]).field)
        except AppRegistryNotReady:
            pass
        return m2m_fields

    def get_prev_record(self):
        """
        Get the previous history record for the instance. `None` if first.
        """
        history = utils.get_history_manager_from_history(self)
        return (
            history.filter(history_date__lt=self.history_date)
            .order_by("history_date")
            .first()
        )


class APISHistoryTableBase(models.Model, GenericModel):
    class Meta:
        abstract = True

    def get_absolute_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:detail", args=[ct, self.history_id])

    def get_reset_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:history:reset", args=[ct, self.history_id])

    def get_diff(self, other_version=None):
        if self.history_type == "-":
            return None
        version = other_version or self.prev_record
        if version:
            delta = self.diff_against(version)
        else:
            delta = self.diff_against(self.__class__())
        changes = list(
            filter(
                lambda x: (x.new != "" or x.old is not None)
                and x.field != "id"
                and not x.field.endswith("_ptr"),
                delta.changes,
            )
        )
        return sorted(changes, key=lambda change: change.field)


class VersionMixin(models.Model):
    history = APISHistoricalRecords(
        inherit=True,
        bases=[
            APISHistoryTableBase,
        ],
        custom_model_name=lambda x: f"Version{x}",
    )
    __history_date = None

    @property
    def _history_date(self):
        return self.__history_date or timezone.now()

    @_history_date.setter
    def _history_date(self, value):
        self.__history_date = value
        pass

    class Meta:
        abstract = True

    def get_history_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:history:history", args=[ct, self.id])

    def _get_historical_relations(self):
        ret = []
        if "apis_core.relations" in settings.INSTALLED_APPS:
            from apis_core.relations.utils import relation_content_types

            ct = ContentType.objects.get_for_model(self)

            rel_content_types = relation_content_types(any_model=type(self))
            rel_models = [ct.model_class() for ct in rel_content_types]
            rel_history_models = [
                model for model in rel_models if issubclass(model, VersionMixin)
            ]

            for model in rel_history_models:
                ret.append(
                    model.history.filter(
                        Q(subj_object_id=self.id, subj_content_type=ct)
                        | Q(obj_object_id=self.id, obj_content_type=ct)
                    ).order_by("history_id")
                )
        return ret

    def get_history_data(self):
        data = []
        prev_entry = None
        queries = self._get_historical_relations()
        flatten_queries = [entry for query in queries for entry in query]

        for entry in flatten_queries:
            if prev_entry is not None:
                if (
                    entry.history_date == prev_entry.history_date
                    and entry.history_user_id == prev_entry.history_user_id
                ):
                    entry.history_type = prev_entry.history_type
                    data[-1] = entry
                    prev_entry = entry
                    continue
            data.append(entry)
            prev_entry = entry
        data += [x for x in self.history.all()]
        data = sorted(data, key=lambda x: x.history_date, reverse=True)
        return data

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



# SPDX-FileCopyrightText: 2023 Birger Schacht
# SPDX-License-Identifier: MIT

import logging
import tomllib
from urllib.parse import urlparse

from django.conf import settings
from django.template.utils import get_app_template_dirs

logger = logging.getLogger(__name__)


def get_entity_settings_by_modelname(entity: str = None) -> dict:
    """
    return the settings for a specific entity or the dict for all entities
    if no entity is given
    """
    apis_entities = getattr(settings, "APIS_ENTITIES", {})
    if entity:
        # lookup entity settings by name and by capitalized name
        return apis_entities.get(entity, apis_entities.get(entity.capitalize(), {}))
    return apis_entities


def dict_from_toml_directory(directory: str) -> dict:
    configs = {}
    pathlists = [path.glob("**/*.toml") for path in get_app_template_dirs(directory)]
    for file in [path for pathlist in pathlists for path in pathlist]:
        try:
            configs[file.resolve()] = tomllib.loads(file.read_text())
        except Exception as e:
            logger.error(f"TOML parser could not read {file}: {e}")
    return configs


def internal_uris() -> list[str]:
    return set([apis_base_uri()] + getattr(settings, "APIS_FORMER_BASE_URIS", []))


def apis_base_uri() -> str:
    return getattr(settings, "APIS_BASE_URI", "https://example.org")


def rdf_namespace_prefix() -> str:
    if hasattr(settings, "APIS_RDF_NAMESPACE_PREFIX"):
        return settings.APIS_RDF_NAMESPACE_PREFIX
    base_uri = urlparse(apis_base_uri())
    return base_uri.hostname.split(".", 1)[0]



import django_filters
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

from apis_core.relations.forms.fields import RelationField
from apis_core.relations.models import Relation


class RelationFilter(django_filters.Filter):
    field_class = RelationField

    def filter(self, qs, value):
        if not value:
            return qs
        relationid, entity_ct_id = value
        subj = Q()
        obj = Q()
        if entity_ct_id:
            content_type_id, entity_id = entity_ct_id.split("_")
            entity = (
                ContentType.objects.get(pk=content_type_id)
                .model_class()
                .objects.get(pk=entity_id)
            )
            subj = Q(subj_content_type=content_type_id, subj_object_id=entity.id)
            obj = Q(obj_content_type=content_type_id, obj_object_id=entity.id)
        relation_model = Relation
        if relationid:
            relation_model = ContentType.objects.get(pk=relationid).model_class()
        obj_rels = relation_model.objects.filter(subj).values_list(
            "obj_object_id", flat=True
        )
        subj_rels = relation_model.objects.filter(obj).values_list(
            "subj_object_id", flat=True
        )
        possible_ids = set(list(subj_rels) + list(obj_rels))
        return qs.filter(pk__in=possible_ids)



# SPDX-FileCopyrightText: 2025 Birger Schacht
# SPDX-License-Identifier: MIT

import logging
import re
import tomllib
from collections import defaultdict
from pathlib import Path

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from django.apps import apps
from django.template.utils import get_app_template_dirs
from rdflib import RDF, BNode, Graph, URIRef

logger = logging.getLogger(__name__)


def resolve(obj, graph):
    """
    Look at the value of object and return the parsed
    value. If the value starts and ens with angle brackets,
    we interpret it as and transform it to an URI.
    If the value is simple text we interpret it as an curie
    and we expand it using the graphs namespace manager.
    Otherwise we simply return the value
    """
    if isinstance(obj, str):
        if obj.startswith("<") and obj.endswith(">"):
            return URIRef(obj[1:-1])
        return graph.namespace_manager.expand_curie(obj)
    if isinstance(obj, bool) and obj is True:
        return None
    return obj


def load_path(path: str | Path) -> dict:
    """
    Load a tomlfile either from a path or from the directory
    `triple_configs` in any of the app directories.
    """
    if isinstance(path, str):
        files = [
            directory / path for directory in get_app_template_dirs("triple_configs")
        ]
        files = list(filter(lambda file: file.exists(), files))
        if files:
            path = files[0]
        else:
            raise ValueError(f"Could not find {path}")
    return tomllib.loads(Path(path).read_text())


def find_regex_matching_configs(uri: str, models: list | None = None) -> dict | None:
    """
    Go through a list of models and return all the rdf configs
    that are configured in those models that have a regex that
    matches the given URI.
    """
    models = models or apps.get_models()
    models_with_config = [model for model in models if hasattr(model, "rdf_configs")]
    configs = []
    for model in models_with_config:
        for regex, path in model.rdf_configs().items():
            if re.match(regex, uri):
                logger.debug(f"{uri} matched {regex}")
                config = load_path(path)
                config["path"] = path
                config["model"] = model
                configs.append(config)
            else:
                logger.debug(f"{uri} did not match {regex}")
    return configs


def find_graph_matching_config(graph: Graph, configs: list[dict] = []) -> dict | None:
    """
    Go through al list of RDF import configs and return the
    ones that have filters defined that match the given graph.
    """
    for config in configs:
        for _filter in config.get("filters", [{None: None}]):
            try:
                triples = []
                for predicate, obj in _filter.items():
                    triples.append(
                        (None, resolve(predicate, graph), resolve(obj, graph))
                    )
                triples = [triple in graph for triple in triples]
                if all(triples):
                    logger.debug("Using %s for parsing graph", config["path"])
                    return config
            except ValueError as e:
                logger.debug("Filter %s does not match: %s", _filter, e)
    return None


def build_sparql_query(curie: str) -> str:
    """
    Build a SPARQL query with language preferences.

    Args:
        curie: predicate to filter on as defined in the toml.
                needs to include the predicate and optionally
                a lang tag to filter for separated with a comma.
                Eg "wdt:P122,en".

    Returns:
        A SPARQL query string
    """
    if curie.lower().strip().startswith(("select", "prefix")):
        return curie
    lang_tag = ""
    if "," in curie:
        curie, lang_tag = curie.split(",", 1)
        lang_tag = f'FILTER LANGMATCHES(LANG(?object), "{lang_tag}")'
    query = f"""
            SELECT ?object 
            WHERE {{ 
                ?subject {curie} ?object {lang_tag}
            }}
        """

    logger.debug("Generated SPARQL query: %s", query)
    return query


def get_value_graph(graph: Graph, curies: str | list[str]) -> list:
    values = []
    if isinstance(curies, str):
        curies = [curies]
    for curie in curies:
        results = graph.query(build_sparql_query(curie))
        objects = [result[0] for result in results]
        for obj in objects:
            if isinstance(obj, BNode):
                values.extend(
                    [
                        str(value)
                        for value in graph.objects(subject=obj)
                        if value != RDF.Seq
                    ]
                )
            else:
                values.append(str(obj))
    return list(dict.fromkeys(values))


def get_something_from_uri(uri: str, models: list | None = None) -> dict | None:
    uri = get_normalized_uri(uri)
    graph = Graph()
    graph.parse(uri)

    configs = find_regex_matching_configs(uri, models)

    if config := find_graph_matching_config(graph, configs):
        result = defaultdict(list)
        result["model"] = config["model"]
        result["relations"] = defaultdict(list)

        for attribute, curies in config.get("attributes", {}).items():
            values = get_value_graph(graph, curies)
            result[attribute].extend(values)
        for relation, details in config.get("relations", {}).items():
            details["curies"] = get_value_graph(graph, details.get("curies", []))
            result["relations"][relation] = details
        return dict(result)
    return None



# Generated by Django 5.1 on 2024-09-18 06:35

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("sample_project", "0002_iscousinof_ispartof_issiblingof_livesin"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="versiongroup",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical group",
                "verbose_name_plural": "historical groups",
            },
        ),
        migrations.AlterModelOptions(
            name="versionperson",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical person",
                "verbose_name_plural": "historical persons",
            },
        ),
        migrations.AlterModelOptions(
            name="versionplace",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical place",
                "verbose_name_plural": "historical places",
            },
        ),
    ]



from apis_core.generic.tables import CustomTemplateColumn, GenericTable


class RelationColumn(CustomTemplateColumn):
    template_name = "columns/relation.html"
    verbose_name = ""


class RelationsListTable(GenericTable):
    relation = RelationColumn()

    class Meta:
        attrs = {"class": "table-sm"}
        sequence = GenericTable.Meta.sequence
        exclude = ("desc",)
